# Copyright 2020-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import contextlib
import datetime
import importlib
import json
import os
import pathlib
import tempfile
import types
from typing import Optional
from unittest.mock import Mock

import craft_parts
import pytest
import responses as responses_module
import yaml
from craft_parts import callbacks, plugins
from craft_providers import Executor, Provider

import charmcraft.parts
from charmcraft import deprecations, instrum, parts
from charmcraft.bases import get_host_as_base
from charmcraft.models import charmcraft as config_module
from charmcraft.models.charmcraft import Base, BasesConfiguration


@pytest.fixture(autouse=True, scope="session")
def tmpdir_under_tmpdir(tmpdir_factory):
    tempfile.tempdir = str(tmpdir_factory.getbasetemp())


@pytest.fixture(autouse=True, scope="session")
def setup_parts():
    parts.setup_parts()


@pytest.fixture()
def new_path(tmp_path):
    old_path = os.getcwd()
    try:
        os.chdir(tmp_path)
        yield tmp_path
    finally:
        os.chdir(old_path)


@pytest.fixture()
def config(tmp_path):
    """Provide a config class with an extra set method for the test to change it."""

    class TestConfig(config_module.CharmcraftConfig, frozen=False):
        """The Config, but with a method to set test values."""

        def set(self, prime=None, **kwargs):
            # prime is special, so we don't need to write all this structure in all tests
            if prime is not None:
                if self.parts is None:
                    self.parts = {}
                self.parts["charm"] = {"plugin": "charm", "prime": prime}

            # the rest is direct
            for k, v in kwargs.items():
                object.__setattr__(self, k, v)

    project = config_module.Project(
        dirpath=tmp_path,
        started_at=datetime.datetime.utcnow(),
        config_provided=True,
    )

    base = BasesConfiguration(**{"build-on": [get_host_as_base()], "run-on": [get_host_as_base()]})

    return TestConfig(
        type="charm",
        bases=[base],
        project=project,
        name="test-charm",
        summary="test summary",
        description="test description",
    )


@pytest.fixture()
def bundle_config(tmp_path):
    """Provide a config class with an extra set method for the test to change it."""

    class TestConfig(config_module.CharmcraftConfig, frozen=False):
        """The Config, but with a method to set test values."""

        def set(self, prime=None, **kwargs):
            # prime is special, so we don't need to write all this structure in all tests
            if prime is not None:
                if self.parts is None:
                    self.parts = {}
                self.parts["bundle"] = {"plugin": "bundle", "prime": prime}

            # the rest is direct
            for k, v in kwargs.items():
                object.__setattr__(self, k, v)

    project = config_module.Project(
        dirpath=tmp_path,
        started_at=datetime.datetime.utcnow(),
        config_provided=True,
    )

    return TestConfig(
        type="bundle",
        project=project,
        name="test-bundle",
        summary="test summary",
        description="test description",
    )


@pytest.fixture(autouse=True)
def intertests_cleanups():
    """Run some cleanups between tests.

    Before each test:

    - reload the instrumentator module to start clean.

    After each test:

    - clear the already-notified structure for each test (this is needed as that
      structure is a module-level one (by design), so otherwise it will be dirty
      between tests).

    - unregister all Craft Parts plugins callbacks
    """
    importlib.reload(instrum)
    yield
    deprecations._ALREADY_NOTIFIED.clear()
    callbacks.unregister_all()


@pytest.fixture()
def responses():
    """Simple helper to use responses module as a fixture, for easier integration in tests."""
    with responses_module.RequestsMock() as rsps:
        yield rsps


@pytest.fixture()
def mock_instance():
    """Provide a mock instance (Executor)."""
    return Mock(spec=Executor)


@pytest.fixture(autouse=True)
def fake_provider(mock_instance):
    """Provide a minimal/fake provider."""

    class FakeProvider(Provider):
        name = "TestProvider"
        install_recommendation = "Insert floppy disk."

        def clean_project_environments(self, *, instance_name: str) -> None:
            pass

        @classmethod
        def ensure_provider_is_available(cls) -> None:
            pass

        def environment(
            self,
            *,
            instance_name: str,
        ) -> Executor:
            return mock_instance

        def create_environment(self, *, instance_name: str):
            yield mock_instance

        @contextlib.contextmanager
        def launched_environment(
            self,
            *,
            project_name: str,
            project_path: pathlib.Path,
            base_configuration: Base,
            instance_name: str,
            allow_unstable: bool = False,
        ):
            yield mock_instance

        @classmethod
        def is_provider_installed(cls) -> bool:
            """Check if provider is installed.

            :returns: True if installed.
            """
            return True

    return FakeProvider()


@pytest.fixture()
def prepare_charmcraft_yaml(tmp_path: pathlib.Path):
    """Helper to create a charmcraft.yaml file in disk.

    If content is not given, remove charmcraft.yaml if exists.
    """

    def prepare_charmcraft_yaml(content: Optional[str] = None):
        if content is None:
            try:
                os.remove(tmp_path / "charmcraft.yaml")
            except OSError:
                pass
        else:
            charmcraft_yaml_file = tmp_path / "charmcraft.yaml"
            charmcraft_yaml_file.write_text(content)

        return tmp_path

    return prepare_charmcraft_yaml


@pytest.fixture()
def prepare_metadata_yaml(tmp_path: pathlib.Path):
    """Helper to create a metadata.yaml file in disk.

    If content is not given, remove metadata.yaml if exists.
    """

    def prepare_metadata_yaml(content: Optional[str] = None, remove: bool = False):
        if content is None:
            try:
                os.remove(tmp_path / "metadata.yaml")
            except OSError:
                pass
        else:
            metadata_yaml_file = tmp_path / "metadata.yaml"
            metadata_yaml_file.write_text(content)

        return tmp_path

    return prepare_metadata_yaml


@pytest.fixture()
def prepare_actions_yaml(tmp_path: pathlib.Path):
    """Helper to create a actions.yaml file in disk.

    If content is not given, remove actions.yaml if exists.
    """

    def prepare_actions_yaml(content: Optional[str] = None):
        if content is None:
            try:
                os.remove(tmp_path / "actions.yaml")
            except OSError:
                pass
        else:
            actions_yaml_file = tmp_path / "actions.yaml"
            actions_yaml_file.write_text(content)

        return tmp_path

    return prepare_actions_yaml


@pytest.fixture()
def prepare_config_yaml(tmp_path: pathlib.Path):
    """Helper to create a config.yaml file in disk.

    If content is not given, remove config.yaml if exists.
    """

    def prepare_config_yaml(content: Optional[str] = None):
        if content is None:
            try:
                os.remove(tmp_path / "config.yaml")
            except OSError:
                pass
        else:
            config_yaml_file = tmp_path / "config.yaml"
            config_yaml_file.write_text(content)

        return tmp_path

    return prepare_config_yaml


@pytest.fixture()
def emitter(emitter):
    """Monkeypatch craft-cli's emitter fixture to easily test the JSON encoded output."""

    def assert_json_output(self, expected_content):
        """Get last output, which should be a message, and validate its content."""
        last_output = self.interactions[-1]
        output_type, raw_output = last_output.args
        assert output_type == "message", "Last command output is not 'message'"
        try:
            output_content = json.loads(raw_output)
        except json.decoder.JSONDecodeError:
            pytest.fail("Last command output is not valid JSON.")
        assert output_content == expected_content

    emitter.assert_json_output = types.MethodType(assert_json_output, emitter)
    return emitter


@pytest.fixture()
def assert_output(capsys):
    """Assert that a given string was sent to stdout.

    This is a helper to simplify tests for charm_builder.py and its modules that print
    directly to stdout.

    Note that every call to this helper will clear the previous captured output.
    """

    def helper(*match_lines):
        captured = capsys.readouterr()
        printed_lines = captured.out.splitlines()
        for match_line in match_lines:
            if match_line not in printed_lines:
                printed_repr = "\n".join(map(repr, printed_lines))
                pytest.fail(f"Line {match_line!r} not found in the output found:\n{printed_repr}")

    return helper


@pytest.fixture()
def build_charm_directory():
    def helper(tmp_path, fake_charms, file_type="charm"):
        expected = {}
        charmcraft_yaml = {"type": file_type}
        for name, path in fake_charms.items():
            full_path = tmp_path / path
            expected[name] = full_path
            full_path.mkdir(parents=True)
            metadata_yaml = {"name": name}
            with (full_path / "charmcraft.yaml").open("w") as yaml_file:
                yaml.safe_dump(charmcraft_yaml, yaml_file)
            with (full_path / "metadata.yaml").open("w") as yaml_file:
                yaml.safe_dump(metadata_yaml, yaml_file)
        return expected

    return helper


@pytest.fixture()
def stub_extensions(monkeypatch):
    from charmcraft.extensions import registry

    extensions_dict = {}
    monkeypatch.setattr(registry, "_EXTENSIONS", extensions_dict)

    return extensions_dict


@pytest.fixture()
def charm_plugin(tmp_path):
    requirement_files = ["reqs1.txt", "reqs2.txt"]
    for req in requirement_files:
        (tmp_path / req).write_text("somedep")
    project_dirs = craft_parts.ProjectDirs(work_dir=tmp_path)
    spec = {
        "plugin": "charm",
        "source": str(tmp_path),
        "charm-entrypoint": "entrypoint",
        "charm-binary-python-packages": ["pkg1", "pkg2"],
        "charm-python-packages": ["pkg3", "pkg4"],
        "charm-requirements": requirement_files,
    }
    plugin_properties = parts.CharmPluginProperties.unmarshal(spec)
    part_spec = plugins.extract_part_properties(spec, plugin_name="charm")
    part = craft_parts.Part(
        "foo", part_spec, project_dirs=project_dirs, plugin_properties=plugin_properties
    )
    project_info = craft_parts.ProjectInfo(
        application_name="test",
        project_dirs=project_dirs,
        cache_dir=tmp_path,
    )
    part_info = craft_parts.PartInfo(project_info=project_info, part=part)

    return plugins.get_plugin(part=part, part_info=part_info, properties=plugin_properties)


@pytest.fixture()
def bundle_plugin(tmp_path):
    project_dirs = craft_parts.ProjectDirs(work_dir=tmp_path)
    spec = {
        "plugin": "bundle",
        "source": str(tmp_path),
    }
    plugin_properties = charmcraft.parts.bundle.BundlePluginProperties.unmarshal(spec)
    part_spec = plugins.extract_part_properties(spec, plugin_name="bundle")
    part = craft_parts.Part(
        "foo", part_spec, project_dirs=project_dirs, plugin_properties=plugin_properties
    )
    project_info = craft_parts.ProjectInfo(
        application_name="test",
        project_dirs=project_dirs,
        cache_dir=tmp_path,
    )
    part_info = craft_parts.PartInfo(project_info=project_info, part=part)

    return plugins.get_plugin(part=part, part_info=part_info, properties=plugin_properties)



# Copyright 2021-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Infrastructure for the 'analyze' command."""
import enum
import pathlib
import tempfile
import textwrap
import zipfile

from craft_cli import CraftError, emit

from charmcraft import linters
from charmcraft.cmdbase import BaseCommand
from charmcraft.models.lint import LintResult
from charmcraft.utils import useful_filepath


class AnalyzeCommand(BaseCommand):
    """Analyze a charm."""

    name = "analyze"
    help_msg = "Analyze a charm"
    overview = textwrap.dedent(
        """
        Analyze a charm.

        Report the attributes and lint results directly in the terminal. Use
        `--force` to run even those configured to be ignored.
    """
    )

    def fill_parser(self, parser):
        """Add own parameters to the general parser."""
        self.include_format_option(parser)
        parser.add_argument("filepath", type=useful_filepath, help="The charm to analyze")
        parser.add_argument(
            "--force",
            action="store_true",
            help="Force to run all checks, even those set to ignore in the configuration",
        )

    def _unzip_charm(self, filepath: pathlib.Path) -> pathlib.Path:
        """Extract the charm content to a temp directory."""
        tmpdir = pathlib.Path(tempfile.mkdtemp())
        try:
            zf = zipfile.ZipFile(str(filepath))
            zf.extractall(path=str(tmpdir))
        except Exception as exc:
            raise CraftError(f"Cannot open charm file {str(filepath)!r}: {exc!r}.")

        # fix permissions as extractall does not keep them (see https://bugs.python.org/issue15795)
        for name in zf.namelist():
            info = zf.getinfo(name)
            inside_zip_mode = info.external_attr >> 16
            extracted_file = tmpdir / name
            current_mode = extracted_file.stat().st_mode
            if current_mode != inside_zip_mode:
                extracted_file.chmod(inside_zip_mode)

        return tmpdir

    def run(self, parsed_args):
        """Run the command."""
        tmpdir = self._unzip_charm(parsed_args.filepath)

        # run the analyzer
        override_ignore_config = bool(parsed_args.force)
        linting_results = linters.analyze(
            self.config,
            tmpdir,
            override_ignore_config=override_ignore_config,
        )

        # if format is json almost no further processing is needed
        if parsed_args.format:
            info = [
                {
                    "name": r.name,
                    "result": r.result,
                    "url": r.url,
                    "type": r.check_type,
                }
                for r in linting_results
            ]
            emit.message(self.format_content(parsed_args.format, info))
            return None

        # group by attributes and lint outcomes (discarding ignored ones)
        grouped = {}
        for result in linting_results:
            if result.check_type == linters.CheckType.ATTRIBUTE:
                group_key = linters.CheckType.ATTRIBUTE
                if isinstance(result.result, enum.Enum):
                    result_info = result.result.value
                else:
                    result_info = result.result
            else:
                # linters
                group_key = result.result
                if result.result == LintResult.OK:
                    result_info = "no issues found"
                elif result.result in (LintResult.FATAL, LintResult.IGNORED):
                    result_info = None
                else:
                    result_info = result.text
            grouped.setdefault(group_key, []).append((result, result_info))

        # present the results
        titles = [
            ("Attributes", linters.CheckType.ATTRIBUTE),
            ("Lint Ignored", LintResult.IGNORED),
            ("Lint Warnings", LintResult.WARNINGS),
            ("Lint Errors", LintResult.ERRORS),
            ("Lint Fatal", LintResult.FATAL),
            ("Lint OK", LintResult.OK),
        ]
        for title, key in titles:
            results = grouped.get(key)
            if results is not None:
                emit.message(f"{title}:")
                for result, result_info in results:
                    if result_info:
                        emit.message(f"- {result.name}: {result_info} ({result.url})")
                    else:
                        emit.message(f"- {result.name} ({result.url})")

        # the return code depends on the presence of different issues
        if LintResult.FATAL in grouped:
            retcode = 1
        elif LintResult.ERRORS in grouped:
            retcode = 2
        elif LintResult.WARNINGS in grouped:
            retcode = 3
        else:
            retcode = 0

        return retcode



# Copyright 2020 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Collection of creation functions for normally used objects for testing."""

import pathlib
import textwrap

from charmcraft.utils import create_importable_name, get_lib_info


def create_lib_filepath(charm_name, lib_name, api=0, patch=1, lib_id="test-lib-id"):
    """Helper to create the structures on disk for a given lib."""
    charm_name = create_importable_name(charm_name)
    base_dir = pathlib.Path("lib")
    lib_file = base_dir / "charms" / charm_name / f"v{api}" / f"{lib_name}.py"
    lib_file.parent.mkdir(parents=True, exist_ok=True)

    # save the content to that specific file under custom structure
    template = textwrap.dedent(
        """
        # test content for a library
        LIBID = "{lib_id}"
        LIBAPI = {api}
        LIBPATCH = {patch}

        # more text and python code...
    """
    )
    content = template.format(lib_id=lib_id, api=api, patch=patch)
    lib_file.write_text(content)

    # use get_lib_info to get the hash of the file, as the used hash is WITHOUT the metadata
    # files (no point in duplicating that logic here)
    libdata = get_lib_info(lib_path=lib_file)
    return content, libdata.content_hash



# Copyright 2021-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Charmcraft environment utilities."""


import distutils.util
import os
import pathlib
from typing import Optional

import platformdirs

from charmcraft.const import SHARED_CACHE_ENV_VAR


def get_host_shared_cache_path():
    """Path for host shared cache."""
    shared_cache_env = os.getenv(SHARED_CACHE_ENV_VAR)
    if shared_cache_env is not None:
        cache_path = pathlib.Path(shared_cache_env).expanduser().resolve()
        cache_path.mkdir(parents=True, exist_ok=True)
        return cache_path

    return platformdirs.user_cache_path(appname="charmcraft", ensure_exists=True)


def get_managed_environment_home_path():
    """Path for home when running in managed environment."""
    return pathlib.Path("/root")


def get_managed_environment_log_path():
    """Path for charmcraft log when running in managed environment."""
    return pathlib.Path("/tmp/charmcraft.log")


def get_managed_environment_metrics_path():
    """Path for charmcraft metrics when running in managed environment."""
    return pathlib.Path("/tmp/metrics.json")


def get_charm_builder_metrics_path():
    """Path for charmcraft metrics when running charm_builder."""
    return pathlib.Path("/tmp/charm_builder_metrics.json")


def get_managed_environment_project_path():
    """Path for project when running in managed environment."""
    return get_managed_environment_home_path() / "project"


def get_managed_environment_snap_channel() -> Optional[str]:
    """User-specified channel to use when installing Charmcraft snap from Snap Store.

    :returns: Channel string if specified, else None.
    """
    return os.getenv("CHARMCRAFT_INSTALL_SNAP_CHANNEL")


def is_charmcraft_running_from_snap():
    """Check if charmcraft is running from the snap."""
    return os.getenv("SNAP_NAME") == "charmcraft" and os.getenv("SNAP") is not None


def is_charmcraft_running_in_developer_mode():
    """Check if Charmcraft is running under developer mode."""
    developer_flag = os.getenv("CHARMCRAFT_DEVELOPER", "n")
    return distutils.util.strtobool(developer_flag) == 1


def is_charmcraft_running_in_managed_mode():
    """Check if charmcraft is running in a managed environment."""
    managed_flag = os.getenv("CHARMCRAFT_MANAGED_MODE", "n")
    return distutils.util.strtobool(managed_flag) == 1



# Copyright 2020-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import sys

import pydantic
import pytest

from charmcraft import parts

pytestmark = pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
# -- tests for part config processing


@pytest.mark.usefixtures("new_path")
def test_partconfig_happy_validation_and_completion():
    data = {
        "plugin": "charm",
        "source": ".",
    }
    completed = parts.process_part_config(data)
    assert completed == {
        "plugin": "charm",
        "source": ".",
        "charm-binary-python-packages": [],
        "charm-entrypoint": "src/charm.py",
        "charm-python-packages": [],
        "charm-requirements": [],
        "charm-strict-dependencies": False,
    }


def test_partconfig_no_plugin():
    data = {
        "source": ".",
    }
    with pytest.raises(ValueError) as raised:
        parts.process_part_config(data)
    assert str(raised.value) == "'plugin' not defined"


def test_partconfig_bad_property():
    data = {
        "plugin": "charm",
        "source": ".",
        "color": "purple",
    }
    with pytest.raises(pydantic.ValidationError) as raised:
        parts.process_part_config(data)
    err = raised.value.errors()
    assert len(err) == 1
    assert err[0]["loc"] == ("color",)
    assert err[0]["msg"] == "extra fields not permitted"


def test_partconfig_bad_type():
    data = {
        "plugin": "charm",
        "source": ["."],
    }
    with pytest.raises(pydantic.ValidationError) as raised:
        parts.process_part_config(data)
    err = raised.value.errors()
    assert len(err) == 2
    assert err[0]["loc"] == ("source",)
    assert err[0]["msg"] == "str type expected"
    assert err[1]["loc"] == ("charm-requirements",)
    assert (
        err[1]["msg"]
        == "cannot validate 'charm-requirements' because invalid 'source' configuration"
    )


def test_partconfig_bad_plugin_property():
    data = {
        "plugin": "charm",
        "charm-timeout": "never",
        "source": ".",
    }
    with pytest.raises(pydantic.ValidationError) as raised:
        parts.process_part_config(data)
    err = raised.value.errors()
    assert len(err) == 1
    assert err[0]["loc"] == ("charm-timeout",)
    assert err[0]["msg"] == "extra fields not permitted"



# Copyright 2020-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Manage the charmcraft tool versioning."""

import fileinput
import pathlib
import subprocess
import sys


def determine_version():
    # Examples (git describe -> python package version):
    # 4.1.1-0-gad012482d -> 4.1.1
    # 4.1.1-16-g2d8943dbc -> 4.1.1.post16+g2d8943dbc
    #
    # For shallow clones or repositories missing tags:
    # 0ae7c04

    desc = (
        subprocess.run(
            ["git", "describe", "--always", "--long"],
            check=True,
            stdout=subprocess.PIPE,
        )
        .stdout.decode()
        .strip()
    )

    split_desc = desc.split("-")
    assert (
        len(split_desc) == 3
    ), f"Failed to parse Charmcraft git version description {desc!r}. Confirm that git repository is present and has the required tags/history."

    version = split_desc[0]
    distance = split_desc[1]
    commit = split_desc[2]

    if distance == "0":
        return version

    return f"{version}.post{distance}+git{commit[1:]}"


def set_charmcraft_iss():
    charmcraft_iss_path = pathlib.Path("windows/charmcraft.iss")
    assert (
        charmcraft_iss_path.exists()
    ), f"Run from project root and ensure {charmcraft_iss_path!s} exists."
    with fileinput.input(str(charmcraft_iss_path), inplace=True) as iss_file:
        for line in iss_file:
            if line.startswith("AppVersion="):
                print(f"AppVersion={determine_version()}")
            else:
                print(line, end="")


if __name__ == "__main__":
    if len(sys.argv) == 2 and sys.argv[1] == "set-charmcraft-iss":
        set_charmcraft_iss()
    else:
        print(determine_version())



# Copyright 2021 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import re
from unittest import mock

import pytest
import snaphelpers

from charmcraft.snap import (
    CharmcraftSnapConfiguration,
    get_snap_configuration,
    validate_snap_configuration,
)


@pytest.fixture()
def mock_snap_config():
    with mock.patch("charmcraft.snap.snaphelpers.SnapConfig", autospec=True) as mock_snap_config:
        yield mock_snap_config


def test_get_snap_configuration_empty(mock_snap_config):
    def fake_get(key: str):
        raise snaphelpers._conf.UnknownConfigKey(key=key)

    mock_snap_config.return_value.get.side_effect = fake_get

    snap_config = get_snap_configuration()

    assert snap_config == CharmcraftSnapConfiguration()


@pytest.mark.parametrize("provider", ["lxd", "multipass"])
def test_get_snap_configuration_valid_providers(mock_snap_config, provider):
    def fake_get(key: str):
        if key == "provider":
            return provider
        raise snaphelpers._conf.UnknownConfigKey(key=key)

    mock_snap_config.return_value.get.side_effect = fake_get

    snap_config = get_snap_configuration()

    assert snap_config == CharmcraftSnapConfiguration(provider=provider)


@pytest.mark.parametrize("provider", ["", "invalid"])
def test_get_snap_configuration_invalid_providers(mock_snap_config, provider):
    def fake_get(key: str):
        if key == "provider":
            return provider
        raise snaphelpers._conf.UnknownConfigKey(key=key)

    mock_snap_config.return_value.get.side_effect = fake_get

    snap_config = get_snap_configuration()

    assert snap_config == CharmcraftSnapConfiguration(provider=provider)
    assert snap_config.provider == provider

    with pytest.raises(ValueError, match=re.escape(f"provider {provider!r} is not supported")):
        validate_snap_configuration(snap_config)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for file-related utilities."""
import os
import pathlib
import sys
import zipfile

import pytest
from craft_cli import CraftError

from charmcraft.utils.file import build_zip, make_executable, useful_filepath


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_make_executable_read_bits(tmp_path):
    pth = tmp_path / "test"
    pth.touch(mode=0o640)
    # validity check
    assert pth.stat().st_mode & 0o777 == 0o640
    with pth.open() as fd:
        make_executable(fd)
        # only read bits got made executable
        assert pth.stat().st_mode & 0o777 == 0o750


def test_usefulfilepath_pathlib(tmp_path):
    """Convert the string to Path."""
    test_file = tmp_path / "testfile.bin"
    test_file.touch()
    path = useful_filepath(str(test_file))
    assert path == test_file
    assert isinstance(path, pathlib.Path)


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_usefulfilepath_home_expanded(tmp_path, monkeypatch):
    """Home-expand the indicated path."""
    fake_home = tmp_path / "homedir"
    fake_home.mkdir()
    test_file = fake_home / "testfile.bin"
    test_file.touch()

    monkeypatch.setitem(os.environ, "HOME", str(fake_home))
    path = useful_filepath("~/testfile.bin")
    assert path == test_file


def test_usefulfilepath_missing():
    """The indicated path is not there."""
    with pytest.raises(CraftError) as cm:
        useful_filepath("not_really_there.txt")
    assert str(cm.value) == "Cannot access 'not_really_there.txt'."


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_usefulfilepath_inaccessible(tmp_path):
    """The indicated path is not readable."""
    test_file = tmp_path / "testfile.bin"
    test_file.touch(mode=0o000)
    with pytest.raises(CraftError) as cm:
        useful_filepath(str(test_file))
    assert str(cm.value) == f"Cannot access {str(test_file)!r}."


def test_usefulfilepath_not_a_file(tmp_path):
    """The indicated path is not a file."""
    with pytest.raises(CraftError) as cm:
        useful_filepath(str(tmp_path))
    assert str(cm.value) == f"{str(tmp_path)!r} is not a file."


def test_zipbuild_simple(tmp_path):
    """Build a bunch of files in the zip."""
    build_dir = tmp_path / "somedir"
    build_dir.mkdir()

    testfile1 = build_dir / "foo.txt"
    testfile1.write_bytes(b"123\x00456")
    subdir = build_dir / "bar"
    subdir.mkdir()
    testfile2 = subdir / "baz.txt"
    testfile2.write_bytes(b"mo\xc3\xb1o")

    zip_filepath = tmp_path / "testresult.zip"
    build_zip(zip_filepath, build_dir)

    zf = zipfile.ZipFile(zip_filepath)
    assert sorted(x.filename for x in zf.infolist()) == ["bar/baz.txt", "foo.txt"]
    assert zf.read("foo.txt") == b"123\x00456"
    assert zf.read("bar/baz.txt") == b"mo\xc3\xb1o"


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_zipbuild_symlink_simple(tmp_path):
    """Symlinks are supported."""
    build_dir = tmp_path / "somedir"
    build_dir.mkdir()

    testfile1 = build_dir / "real.txt"
    testfile1.write_bytes(b"123\x00456")
    testfile2 = build_dir / "link.txt"
    testfile2.symlink_to(testfile1)

    zip_filepath = tmp_path / "testresult.zip"
    build_zip(zip_filepath, build_dir)

    zf = zipfile.ZipFile(zip_filepath)
    assert sorted(x.filename for x in zf.infolist()) == ["link.txt", "real.txt"]
    assert zf.read("real.txt") == b"123\x00456"
    assert zf.read("link.txt") == b"123\x00456"


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_zipbuild_symlink_outside(tmp_path):
    """No matter where the symlink points to."""
    # outside the build dir
    testfile1 = tmp_path / "real.txt"
    testfile1.write_bytes(b"123\x00456")

    # inside the build dir
    build_dir = tmp_path / "somedir"
    build_dir.mkdir()
    testfile2 = build_dir / "link.txt"
    testfile2.symlink_to(testfile1)

    zip_filepath = tmp_path / "testresult.zip"
    build_zip(zip_filepath, build_dir)

    zf = zipfile.ZipFile(zip_filepath)
    assert sorted(x.filename for x in zf.infolist()) == ["link.txt"]
    assert zf.read("link.txt") == b"123\x00456"



# Copyright 2020-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

from argparse import Namespace

from charmcraft import __version__
from charmcraft.commands.version import VersionCommand


def test_version_result(emitter):
    """Check it produces the right version."""
    cmd = VersionCommand("config")
    args = Namespace(format=None)
    cmd.run(args)
    emitter.assert_message(__version__)


def test_version_result_formatjson(emitter):
    """Format the output."""
    cmd = VersionCommand("config")
    args = Namespace(format="json")
    cmd.run(args)
    emitter.assert_json_output({"version": __version__})



# Copyright 2020-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import datetime
import os
import pathlib
import sys
from textwrap import dedent
from unittest.mock import patch

import pytest
from craft_cli import CraftError

from charmcraft import linters
from charmcraft.config import load
from charmcraft.models.charmcraft import Base, BasesConfiguration, CharmhubConfig
from charmcraft.utils import get_host_architecture

# -- tests for the config loading


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                """
            ),
            None,
        ],
    ],
)
def test_load_current_directory(
    tmp_path,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml,
    metadata_yaml,
    monkeypatch,
):
    """Init the config using charmcraft.yaml in current directory."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    monkeypatch.chdir(tmp_path)
    fake_utcnow = datetime.datetime(1970, 1, 1, 0, 0, 2, tzinfo=datetime.timezone.utc)
    with patch("datetime.datetime") as mock:
        mock.utcnow.return_value = fake_utcnow
        config = load(None)
    assert config.type == "charm"
    assert config.project.dirpath == tmp_path
    assert config.project.config_provided
    assert config.project.started_at == fake_utcnow


def test_load_managed_mode_directory(monkeypatch, tmp_path):
    """Validate managed-mode default directory is /root/project."""
    monkeypatch.chdir(tmp_path)
    monkeypatch.setenv("CHARMCRAFT_MANAGED_MODE", "1")

    # Patch out Config (and Project) to prevent directory validation checks.
    with patch("charmcraft.config.CharmcraftConfig"):
        with patch("charmcraft.config.Project") as mock_project:
            with patch("charmcraft.config.load_yaml"):
                load(None)

    assert mock_project.call_args.kwargs["dirpath"] == pathlib.Path("/root/project")


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                """
            ),
            None,
        ],
    ],
)
def test_load_specific_directory_ok(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Init the config using charmcraft.yaml in a specific directory."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(tmp_path)
    assert config.type == "charm"
    assert config.project.dirpath == tmp_path


def test_load_optional_charmcraft_missing(tmp_path):
    """Specify a directory where the file is missing."""
    config = load(tmp_path)
    assert config.project.dirpath == tmp_path
    assert not config.project.config_provided


def test_load_optional_charmcraft_bad_directory(tmp_path):
    """Specify a missing directory."""
    missing_directory = tmp_path / "missing"
    config = load(missing_directory)
    assert config.project.dirpath == missing_directory
    assert not config.project.config_provided


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                """
            ),
            None,
        ],
    ],
)
def test_load_specific_directory_resolved(
    tmp_path,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml,
    metadata_yaml,
    monkeypatch,
):
    """Ensure that the given directory is resolved to always show the whole path."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    # change to some dir, and reference the config dir relatively
    subdir = tmp_path / "subdir"
    subdir.mkdir()
    monkeypatch.chdir(subdir)
    config = load("../")

    assert config.type == "charm"
    assert config.project.dirpath == tmp_path


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                """
            ),
            None,
        ],
    ],
)
@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_load_specific_directory_expanded(
    tmp_path,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml,
    metadata_yaml,
    monkeypatch,
):
    """Ensure that the given directory is user-expanded."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    # fake HOME so the '~' indication is verified to work
    monkeypatch.setitem(os.environ, "HOME", str(tmp_path))
    config = load("~")

    assert config.type == "charm"
    assert config.project.dirpath == tmp_path


def test_load_metadata_keys_exists_both(tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml):
    """Cannot define metadata keys in both charmcraft.yaml and metadata.yaml."""
    prepare_charmcraft_yaml(
        dedent(
            """\
            name: test-charm-name
            type: charm
            """
        )
    )
    prepare_metadata_yaml(
        dedent(
            """\
            name: test-charm-name
            summary: test-summary
            description: test-description
            """
        )
    )

    with pytest.raises(CraftError) as cm:
        load(tmp_path)

    assert str(cm.value) == dedent(
        """Cannot specify 'name' in charmcraft.yaml when 'metadata.yaml' exists"""
    )


# -- tests for schema restrictions


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                whatever: new-stuff
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                whatever: new-stuff
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                """
            ),
            None,
        ],
    ],
)
def test_schema_top_level_no_extra_properties(
    tmp_path,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml,
    metadata_yaml,
):
    """Schema validation, cannot add undefined properties at the top level."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - extra field 'whatever' not permitted in top-level configuration"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                charmhub:
                  api-url: https://www.canonical.com
                bases:
                  - build-on:
                      - name: test-build-name
                        channel: test-build-channel
                    run-on:
                      - name: test-build-name
                        channel: test-build-channel
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                charmhub:
                  api-url: https://www.canonical.com
                bases:
                  - build-on:
                      - name: test-build-name
                        channel: test-build-channel
                    run-on:
                      - name: test-build-name
                        channel: test-build-channel
                """
            ),
            None,
        ],
    ],
)
def test_schema_type_missing(
    tmp_path,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml,
    metadata_yaml,
):
    """Schema validation, type is mandatory."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - field 'type' required in top-level configuration"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: 33
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: 33
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                """
            ),
            None,
        ],
    ],
)
def test_schema_type_bad_type(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Schema validation, type is a string."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - unexpected value; permitted: 'bundle', 'charm' in field 'type'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: whatever
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: whatever
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                """
            ),
            None,
        ],
    ],
)
def test_schema_type_limited_values(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Schema validation, type must be a subset of values."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - unexpected value; permitted: 'bundle', 'charm' in field 'type'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                charmhub:
                  api-url: 33
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                charmhub:
                  api-url: 33
                """
            ),
            None,
        ],
    ],
)
def test_schema_charmhub_api_url_bad_type(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Schema validation, charmhub.api-url must be a string."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - invalid or missing URL scheme in field 'charmhub.api-url'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                charmhub:
                  api-url: stuff.com
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                charmhub:
                  api-url: stuff.com
                """
            ),
            None,
        ],
    ],
)
def test_schema_charmhub_api_url_bad_format(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Schema validation, charmhub.api-url must be a full URL."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - invalid or missing URL scheme in field 'charmhub.api-url'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                charmhub:
                  storage-url: 33
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                charmhub:
                  storage-url: 33
                """
            ),
            None,
        ],
    ],
)
def test_schema_charmhub_storage_url_bad_type(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Schema validation, charmhub.storage-url must be a string."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - invalid or missing URL scheme in field 'charmhub.storage-url'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                charmhub:
                  storage-url: stuff.com
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                charmhub:
                  storage-url: stuff.com
                """
            ),
            None,
        ],
    ],
)
def test_schema_charmhub_storage_url_bad_format(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Schema validation, charmhub.storage-url must be a full URL."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - invalid or missing URL scheme in field 'charmhub.storage-url'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                charmhub:
                  registry-url: 33
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                charmhub:
                  registry-url: 33
                """
            ),
            None,
        ],
    ],
)
def test_schema_charmhub_registry_url_bad_type(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Schema validation, charmhub.registry-url must be a string."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - invalid or missing URL scheme in field 'charmhub.registry-url'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                charmhub:
                  registry-url: stuff.com
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                charmhub:
                  registry-url: stuff.com
                """
            ),
            None,
        ],
    ],
)
def test_schema_charmhub_registry_url_bad_format(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Schema validation, charmhub.registry-url must be a full URL."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - invalid or missing URL scheme in field 'charmhub.registry-url'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                charmhub:
                  storage-url: https://some.server.com
                  crazy: false
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                charmhub:
                  storage-url: https://some.server.com
                  crazy: false
                """
            ),
            None,
        ],
    ],
)
def test_schema_charmhub_no_extra_properties(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Schema validation, cannot add undefined properties in charmhub key."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - extra field 'crazy' not permitted in 'charmhub' configuration"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: bundle
                parts: ['foo', 'bar']
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: bundle
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                parts: ['foo', 'bar']
                """
            ),
            None,
        ],
    ],
)
def test_schema_basicprime_bad_init_structure(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Schema validation, basic prime with bad parts."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - value must be a dictionary in field 'parts'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: bundle
                parts:
                  charm: ['foo', 'bar']
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: bundle
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                parts:
                  charm: ['foo', 'bar']
                """
            ),
            None,
        ],
    ],
)
def test_schema_basicprime_bad_bundle_structure(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Schema validation, basic prime with bad bundle."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - part 'charm' must be a dictionary in field 'parts'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: bundle
                parts:
                  charm:
                    prime: foo
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: bundle
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                parts:
                  charm:
                    prime: foo
                """
            ),
            None,
        ],
    ],
)
def test_schema_basicprime_bad_prime_structure(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Schema validation, basic prime with bad prime."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - value is not a valid list in field 'parts.charm.prime'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: bundle
                parts:
                  charm:
                    prime: [{}, 'foo']
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: bundle
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                parts:
                  charm:
                    prime: [{}, 'foo']
                """
            ),
            None,
        ],
    ],
)
def test_schema_basicprime_bad_prime_type(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Schema validation, basic prime with a prime holding not strings."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - string type expected in field 'parts.charm.prime[0]'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: bundle
                parts:
                  charm:
                    prime: ['', 'foo']
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: bundle
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description

                parts:
                  charm:
                    prime: ['', 'foo']
                """
            ),
            None,
        ],
    ],
)
def test_schema_basicprime_bad_prime_type_empty(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Schema validation, basic prime with a prime holding not strings."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - path cannot be empty in field 'parts.charm.prime[0]'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: bundle
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                parts:
                  charm:
                    prime: ['/bar/foo', 'foo']
                """
            ),
            None,
        ],
    ],
)
def test_schema_basicprime_bad_content_format(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Schema validation, basic prime with a prime holding not strings."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - '/bar/foo' must be a relative path (cannot start with '/') in field 'parts.charm.prime[0]'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: bundle
                parts:
                  other-part: 1
                  charm:
                    prime: ['/bar/foo', 'foo']
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: bundle
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                parts:
                  other-part: 1
                  charm:
                    prime: ['/bar/foo', 'foo']
                """
            ),
            None,
        ],
    ],
)
def test_schema_additional_part(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Schema validation, basic prime with bad part."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - part 'other-part' must be a dictionary in field 'parts'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:  # mandatory
                  - build-on:
                      - name: test-build-name
                        channel: test-build-channel
                    run-on:
                      - name: test-build-name
                        channel: test-build-channel
                parts:
                  other-part:
                    plugin: charm
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                bases:  # mandatory
                  - build-on:
                      - name: test-build-name
                        channel: test-build-channel
                    run-on:
                      - name: test-build-name
                        channel: test-build-channel
                parts:
                  other-part:
                    plugin: charm
                """
            ),
            None,
        ],
    ],
)
def test_schema_other_charm_part_no_source(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Schema validation, basic prime with bad part."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - field 'source' required in 'parts.other-part' configuration
        - cannot validate 'charm-requirements' because invalid 'source' configuration in field 'parts.other-part.charm-requirements'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: bundle
                parts:
                  other-part:
                    plugin: bundle
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: bundle
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                parts:
                  other-part:
                    plugin: bundle
                """
            ),
            None,
        ],
    ],
)
def test_schema_other_bundle_part_no_source(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Schema validation, basic prime with bad part."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - field 'source' required in 'parts.other-part' configuration"""
    )


# -- tests to check the double layer schema loading; using the 'charm' plugin
#    because it is the default (and has good default properties to be overridden and )
#    the 'dump' one because it's a special case of no having a model


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - name: somebase
                    channel: "30.04"
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                bases:
                  - name: somebase
                    channel: "30.04"
                """
            ),
            None,
        ],
    ],
)
def test_schema_doublelayer_no_parts_type_charm(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """No 'parts' specified at all, full default to charm plugin."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(tmp_path)
    assert config.parts == {
        "charm": {
            "plugin": "charm",
            "source": str(tmp_path),
            "charm-binary-python-packages": [],
            "charm-entrypoint": "src/charm.py",
            "charm-python-packages": [],
            "charm-requirements": [],
            "charm-strict-dependencies": False,
        }
    }


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: bundle
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: bundle
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                """
            ),
            None,
        ],
    ],
)
def test_schema_doublelayer_no_parts_type_bundle(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """No 'parts' specified at all, full default to bundle plugin."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(tmp_path)
    assert config.parts == {
        "bundle": {
            "plugin": "bundle",
            "source": str(tmp_path),
        }
    }


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - name: somebase
                    channel: "30.04"
                parts:
                  mycharm:
                    plugin: dump
                    source: https://the.net/whatever.tar.gz
                    source-type: tar
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                bases:
                  - name: somebase
                    channel: "30.04"
                parts:
                  mycharm:
                    plugin: dump
                    source: https://the.net/whatever.tar.gz
                    source-type: tar
                """
            ),
            None,
        ],
    ],
)
def test_schema_doublelayer_parts_no_charm(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """The 'parts' key is specified, but no 'charm' entry."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(tmp_path)
    assert config.parts == {
        "mycharm": {
            "plugin": "dump",
            "source": "https://the.net/whatever.tar.gz",
            "source-type": "tar",
        }
    }


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - name: somebase
                    channel: "30.04"
                parts:
                  charm:
                    prime: [to_be_included.*]  # random key to have a valid yaml
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                bases:
                  - name: somebase
                    channel: "30.04"
                parts:
                  charm:
                    prime: [to_be_included.*]  # random key to have a valid yaml
                """
            ),
            None,
        ],
    ],
)
def test_schema_doublelayer_parts_with_charm_plugin_missing(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """A charm part is specified but no plugin is indicated."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(tmp_path)
    assert config.parts == {
        "charm": {
            "plugin": "charm",
            "source": str(tmp_path),
            "charm-binary-python-packages": [],
            "charm-entrypoint": "src/charm.py",
            "charm-python-packages": [],
            "charm-requirements": [],
            "prime": ["to_be_included.*"],
            "charm-strict-dependencies": False,
        }
    }


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - name: somebase
                    channel: "30.04"
                parts:
                  charm:
                    plugin: charm
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                bases:
                  - name: somebase
                    channel: "30.04"
                parts:
                  charm:
                    plugin: charm
                """
            ),
            None,
        ],
    ],
)
def test_schema_doublelayer_parts_with_charm_plugin_charm(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """A charm part is fully specified."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(tmp_path)
    assert config.parts == {
        "charm": {
            "plugin": "charm",
            "source": str(tmp_path),
            "charm-binary-python-packages": [],
            "charm-entrypoint": "src/charm.py",
            "charm-python-packages": [],
            "charm-requirements": [],
            "charm-strict-dependencies": False,
        }
    }


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - name: somebase
                    channel: "30.04"
                parts:
                  charm:
                    plugin: dump
                    source: https://the.net/whatever.tar.gz
                    source-type: tar
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                bases:
                  - name: somebase
                    channel: "30.04"
                parts:
                  charm:
                    plugin: dump
                    source: https://the.net/whatever.tar.gz
                    source-type: tar
                """
            ),
            None,
        ],
    ],
)
def test_schema_doublelayer_parts_with_charm_plugin_different(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """There is a 'charm' part but using a different plugin."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(tmp_path)
    assert config.parts == {
        "charm": {
            "plugin": "dump",
            "source": "https://the.net/whatever.tar.gz",
            "source-type": "tar",
        }
    }


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - name: somebase
                    channel: "30.04"
                parts:
                  charm:
                    charm-entrypoint: different.py
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                bases:
                  - name: somebase
                    channel: "30.04"
                parts:
                  charm:
                    charm-entrypoint: different.py
                """
            ),
            None,
        ],
    ],
)
def test_schema_doublelayer_parts_with_charm_overriding_properties(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """A charm plugin is used and its default properties are overridden."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(tmp_path)
    assert config.parts == {
        "charm": {
            "plugin": "charm",
            "source": str(tmp_path),
            "charm-binary-python-packages": [],
            "charm-entrypoint": "different.py",
            "charm-python-packages": [],
            "charm-requirements": [],
            "charm-strict-dependencies": False,
        }
    }


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - name: somebase
                    channel: "30.04"
                parts:
                  charm:
                    charm-point: different.py  # misspelled!
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                bases:
                  - name: somebase
                    channel: "30.04"
                parts:
                  charm:
                    charm-point: different.py  # misspelled!
                """
            ),
            None,
        ],
    ],
)
def test_schema_doublelayer_parts_with_charm_validating_props(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """A charm plugin is used and its validation schema is triggered ok."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - extra field 'charm-point' not permitted in 'parts.charm' configuration"""
    )


# -- tests for Charmhub config


def test_charmhub_frozen():
    """Cannot change values from the charmhub config."""
    config = CharmhubConfig()
    with pytest.raises(TypeError):
        config.api_url = "broken"


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: bundle
                charmhub:
                  storage_url: https://server1.com
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: bundle
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                charmhub:
                  storage_url: https://server1.com
                """
            ),
            None,
        ],
    ],
)
def test_charmhub_underscore_in_names(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Do not support underscore in attributes, only dash."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - extra field 'storage_url' not permitted in 'charmhub' configuration"""
    )


# -- tests for bases


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: bundle
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: bundle
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                """
            ),
            None,
        ],
    ],
)
def test_no_bases_is_ok_for_bundles(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Do not send a deprecation message if it is a bundle."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(tmp_path)
    assert config.type == "bundle"


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: bundle
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                bases:
                  - build-on:
                      - name: test-build-name
                        channel: test-build-channel
                """
            ),
            None,
        ],
    ],
)
def test_bases_forbidden_for_bundles(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Do not allow a bases configuration for bundles."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - Field not allowed when type=bundle in field 'bases'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - build-on:
                      - name: test-build-name
                        channel: test-build-channel
                    run-on:
                      - name: test-run-name
                        channel: test-run-channel
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                bases:
                  - build-on:
                      - name: test-build-name
                        channel: test-build-channel
                    run-on:
                      - name: test-run-name
                        channel: test-run-channel
                """
            ),
            None,
        ],
    ],
)
def test_bases_minimal_long_form(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Minimal bases configuration, long form."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(tmp_path)
    assert config.bases == [
        BasesConfiguration(
            **{
                "build-on": [
                    Base(
                        name="test-build-name",
                        channel="test-build-channel",
                        architectures=[get_host_architecture()],
                    ),
                ],
                "run-on": [
                    Base(
                        name="test-run-name",
                        channel="test-run-channel",
                        architectures=[get_host_architecture()],
                    ),
                ],
            }
        )
    ]


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - build-on:
                      - name: test-name
                        channel: test-build-channel
                        extra-extra: read all about it
                    run-on:
                      - name: test-name
                        channel: test-run-channel
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                bases:
                  - build-on:
                      - name: test-name
                        channel: test-build-channel
                        extra-extra: read all about it
                    run-on:
                      - name: test-name
                        channel: test-run-channel
                """
            ),
            None,
        ],
    ],
)
def test_bases_extra_field_error(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Extra field in bases configuration."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - extra field 'extra-extra' not permitted in 'bases[0].build-on[0]' configuration"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                bases:
                  - build_on:
                      - name: test-name
                        channel: test-build-channel
                    run_on:
                      - name: test-name
                        channel: test-run-channel
                """
            ),
            None,
        ],
    ],
)
def test_bases_underscores_error(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - field 'build-on' required in 'bases[0]' configuration
        - field 'run-on' required in 'bases[0]' configuration
        - extra field 'build_on' not permitted in 'bases[0]' configuration
        - extra field 'run_on' not permitted in 'bases[0]' configuration"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - build-on:
                      - name: test-build-name
                        channel: 20.10
                    run-on:
                      - name: test-run-name
                        channel: test-run-channel
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                bases:
                  - build-on:
                      - name: test-build-name
                        channel: 20.10
                    run-on:
                      - name: test-run-name
                        channel: test-run-channel
                """
            ),
            None,
        ],
    ],
)
def test_channel_is_yaml_number(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - string type expected in field 'bases[0].build-on[0].channel'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - build-on:
                      - name: test-build-name
                        channel: test-build-channel
                    run-on:
                      - name: test-run-name
                        channel: test-run-channel
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                bases:
                  - build-on:
                      - name: test-build-name
                        channel: test-build-channel
                    run-on:
                      - name: test-run-name
                        channel: test-run-channel
                """
            ),
            None,
        ],
    ],
)
def test_minimal_long_form_bases(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(tmp_path)
    assert config.bases == [
        BasesConfiguration(
            **{
                "build-on": [
                    Base(
                        name="test-build-name",
                        channel="test-build-channel",
                        architectures=[get_host_architecture()],
                    ),
                ],
                "run-on": [
                    Base(
                        name="test-run-name",
                        channel="test-run-channel",
                        architectures=[get_host_architecture()],
                    ),
                ],
            }
        )
    ]


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - build-on:
                      - name: test-build-name-1
                        channel: test-build-channel-1
                      - name: test-build-name-2
                        channel: test-build-channel-2
                      - name: test-build-name-3
                        channel: test-build-channel-3
                        architectures: [riscVI]
                    run-on:
                      - name: test-run-name-1
                        channel: test-run-channel-1
                        architectures: [amd64]
                      - name: test-run-name-2
                        channel: test-run-channel-2
                        architectures: [amd64, arm64]
                      - name: test-run-name-3
                        channel: test-run-channel-3
                        architectures: [amd64, arm64, riscVI]
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                bases:
                  - build-on:
                      - name: test-build-name-1
                        channel: test-build-channel-1
                      - name: test-build-name-2
                        channel: test-build-channel-2
                      - name: test-build-name-3
                        channel: test-build-channel-3
                        architectures: [riscVI]
                    run-on:
                      - name: test-run-name-1
                        channel: test-run-channel-1
                        architectures: [amd64]
                      - name: test-run-name-2
                        channel: test-run-channel-2
                        architectures: [amd64, arm64]
                      - name: test-run-name-3
                        channel: test-run-channel-3
                        architectures: [amd64, arm64, riscVI]
                """
            ),
            None,
        ],
    ],
)
def test_complex_long_form_bases(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(tmp_path)
    assert config.bases == [
        BasesConfiguration(
            **{
                "build-on": [
                    Base(
                        name="test-build-name-1",
                        channel="test-build-channel-1",
                        architectures=[get_host_architecture()],
                    ),
                    Base(
                        name="test-build-name-2",
                        channel="test-build-channel-2",
                        architectures=[get_host_architecture()],
                    ),
                    Base(
                        name="test-build-name-3",
                        channel="test-build-channel-3",
                        architectures=["riscVI"],
                    ),
                ],
                "run-on": [
                    Base(
                        name="test-run-name-1",
                        channel="test-run-channel-1",
                        architectures=["amd64"],
                    ),
                    Base(
                        name="test-run-name-2",
                        channel="test-run-channel-2",
                        architectures=["amd64", "arm64"],
                    ),
                    Base(
                        name="test-run-name-3",
                        channel="test-run-channel-3",
                        architectures=["amd64", "arm64", "riscVI"],
                    ),
                ],
            }
        )
    ]


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - build-on:
                      - name: test-build-name-1
                        channel: test-build-channel-1
                    run-on:
                      - name: test-run-name-1
                        channel: test-run-channel-1
                        architectures: [amd64, arm64]
                  - build-on:
                      - name: test-build-name-2
                        channel: test-build-channel-2
                    run-on:
                      - name: test-run-name-2
                        channel: test-run-channel-2
                        architectures: [amd64, arm64]
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                bases:
                  - build-on:
                      - name: test-build-name-1
                        channel: test-build-channel-1
                    run-on:
                      - name: test-run-name-1
                        channel: test-run-channel-1
                        architectures: [amd64, arm64]
                  - build-on:
                      - name: test-build-name-2
                        channel: test-build-channel-2
                    run-on:
                      - name: test-run-name-2
                        channel: test-run-channel-2
                        architectures: [amd64, arm64]
                """
            ),
            None,
        ],
    ],
)
def test_multiple_long_form_bases(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(tmp_path)
    assert config.bases == [
        BasesConfiguration(
            **{
                "build-on": [
                    Base(
                        name="test-build-name-1",
                        channel="test-build-channel-1",
                        architectures=[get_host_architecture()],
                    ),
                ],
                "run-on": [
                    Base(
                        name="test-run-name-1",
                        channel="test-run-channel-1",
                        architectures=["amd64", "arm64"],
                    ),
                ],
            }
        ),
        BasesConfiguration(
            **{
                "build-on": [
                    Base(
                        name="test-build-name-2",
                        channel="test-build-channel-2",
                        architectures=[get_host_architecture()],
                    ),
                ],
                "run-on": [
                    Base(
                        name="test-run-name-2",
                        channel="test-run-channel-2",
                        architectures=["amd64", "arm64"],
                    ),
                ],
            }
        ),
    ]


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - name: test-name
                    channel: test-channel
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                bases:
                  - name: test-name
                    channel: test-channel
                """
            ),
            None,
        ],
    ],
)
def test_bases_minimal_short_form(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(tmp_path)
    assert config.bases == [
        BasesConfiguration(
            **{
                "build-on": [
                    Base(
                        name="test-name",
                        channel="test-channel",
                        architectures=[get_host_architecture()],
                    ),
                ],
                "run-on": [
                    Base(
                        name="test-name",
                        channel="test-channel",
                        architectures=[get_host_architecture()],
                    ),
                ],
            }
        )
    ]


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - name: test-name
                    channel: test-channel
                    extra-extra: read all about it
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
    ],
)
def test_bases_short_form_extra_field_error(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - extra field 'extra-extra' not permitted in 'bases[0]' configuration"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - name: test-name
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                bases:
                  - name: test-name
                """
            ),
            None,
        ],
    ],
)
def test_bases_short_form_missing_field_error(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - field 'channel' required in 'bases[0]' configuration"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                bases:
                  - name: test-name
                  - build-on:
                      - name: test-build-name
                    run-on:
                      - name: test-run-name
                """
            ),
            None,
        ],
    ],
)
def test_bases_mixed_form_errors(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Only the short-form errors are exposed as its the first validation pass."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - field 'channel' required in 'bases[0]' configuration"""
    )


# -- tests for analysis


@pytest.fixture()
def create_checker(monkeypatch):
    """Helper to patch and add checkers to the real structure."""
    test_checkers = []
    monkeypatch.setattr(linters, "CHECKERS", test_checkers)

    def add_checker(c_name, c_type):
        class FakeChecker:
            name = c_name
            check_type = c_type

        test_checkers.append(FakeChecker)

    return add_checker


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: bundle
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: bundle
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                """
            ),
            None,
        ],
    ],
)
def test_schema_analysis_missing(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """No analysis configuration leads to some defaults in place."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(tmp_path)
    assert config.analysis.ignore.attributes == []
    assert config.analysis.ignore.linters == []


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: bundle
                analysis:
                  ignore:
                    attributes: []
                    linters: []
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: bundle
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                analysis:
                  ignore:
                    attributes: []
                    linters: []
                """
            ),
            None,
        ],
    ],
)
def test_schema_analysis_full_struct_just_empty(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """Complete analysis structure, empty."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(tmp_path)
    assert config.analysis.ignore.attributes == []
    assert config.analysis.ignore.linters == []


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: bundle
                analysis:
                  ignore:
                    attributes: [check_ok_1, check_ok_2]
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: bundle
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                analysis:
                  ignore:
                    attributes: [check_ok_1, check_ok_2]
                """
            ),
            None,
        ],
    ],
)
def test_schema_analysis_ignore_attributes(
    tmp_path,
    create_checker,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml,
    metadata_yaml,
):
    """Some attributes are correctly ignored."""
    create_checker("check_ok_1", linters.CheckType.ATTRIBUTE)
    create_checker("check_ok_2", linters.CheckType.ATTRIBUTE)
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(tmp_path)
    assert config.analysis.ignore.attributes == ["check_ok_1", "check_ok_2"]
    assert config.analysis.ignore.linters == []


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: bundle
                analysis:
                  ignore:
                    linters: [check_ok_1, check_ok_2]
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: bundle
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                analysis:
                  ignore:
                    linters: [check_ok_1, check_ok_2]
                """
            ),
            None,
        ],
    ],
)
def test_schema_analysis_ignore_linters(
    tmp_path,
    create_checker,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml,
    metadata_yaml,
):
    """Some linters are correctly ignored."""
    create_checker("check_ok_1", linters.CheckType.LINT)
    create_checker("check_ok_2", linters.CheckType.LINT)
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(tmp_path)
    assert config.analysis.ignore.attributes == []
    assert config.analysis.ignore.linters == ["check_ok_1", "check_ok_2"]


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: bundle
                analysis:
                  ignore:
                    attributes: [check_ok_1, check_missing]
                    linters: [check_ok_2]
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: bundle
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                analysis:
                  ignore:
                    attributes: [check_ok_1, check_missing]
                    linters: [check_ok_2]
                """
            ),
            None,
        ],
    ],
)
def test_schema_analysis_ignore_attribute_missing(
    tmp_path,
    create_checker,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml,
    metadata_yaml,
):
    """An attribute specified to ignore is missing in the system."""
    create_checker("check_ok_1", linters.CheckType.ATTRIBUTE)
    create_checker("check_ok_2", linters.CheckType.LINT)
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - Bad attribute name 'check_missing' in field 'analysis.ignore.attributes[1]'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: bundle
                analysis:
                  ignore:
                    attributes: [check_ok_1]
                    linters: [check_ok_2, check_missing]
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: bundle
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                analysis:
                  ignore:
                    attributes: [check_ok_1]
                    linters: [check_ok_2, check_missing]
                """
            ),
            None,
        ],
    ],
)
def test_schema_analysis_ignore_linter_missing(
    tmp_path,
    create_checker,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml,
    metadata_yaml,
):
    """A linter specified to ignore is missing in the system."""
    create_checker("check_ok_1", linters.CheckType.ATTRIBUTE)
    create_checker("check_ok_2", linters.CheckType.LINT)
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError) as cm:
        load(tmp_path)
    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - Bad lint name 'check_missing' in field 'analysis.ignore.linters[1]'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - name: test-name
                    channel: test-channel
                actions:
                  pause:
                    description: Pause the database.
                  resume:
                    description: Resume a paused database.
                  snapshot:
                    description: Take a snapshot of the database.
                    params:
                      filename:
                        type: string
                        description: The name of the snapshot file.
                      compression:
                        type: object
                        description: The type of compression to use.
                        properties:
                          kind:
                            type: string
                            enum: [gzip, bzip2, xz]
                          quality:
                            description: Compression quality
                            type: integer
                            minimum: 0
                            maximum: 9
                    required: [filename]
                    additionalProperties: false
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                bases:
                  - name: test-name
                    channel: test-channel
                actions:
                  pause:
                    description: Pause the database.
                  resume:
                    description: Resume a paused database.
                  snapshot:
                    description: Take a snapshot of the database.
                    params:
                      filename:
                        type: string
                        description: The name of the snapshot file.
                      compression:
                        type: object
                        description: The type of compression to use.
                        properties:
                          kind:
                            type: string
                            enum: [gzip, bzip2, xz]
                          quality:
                            description: Compression quality
                            type: integer
                            minimum: 0
                            maximum: 9
                    required: [filename]
                    additionalProperties: false
                """
            ),
            None,
        ],
    ],
)
def test_actions_defined_in_charmcraft_yaml(
    tmp_path, prepare_charmcraft_yaml, prepare_metadata_yaml, charmcraft_yaml, metadata_yaml
):
    """test actions defined in charmcraft.yaml"""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)

    config = load(tmp_path)

    assert config.actions.actions == {
        "pause": {"description": "Pause the database."},
        "resume": {"description": "Resume a paused database."},
        "snapshot": {
            "description": "Take a snapshot of the database.",
            "params": {
                "filename": {"type": "string", "description": "The name of the snapshot file."},
                "compression": {
                    "type": "object",
                    "description": "The type of compression to use.",
                    "properties": {
                        "kind": {"type": "string", "enum": ["gzip", "bzip2", "xz"]},
                        "quality": {
                            "description": "Compression quality",
                            "type": "integer",
                            "minimum": 0,
                            "maximum": 9,
                        },
                    },
                },
            },
            "required": ["filename"],
            "additionalProperties": False,
        },
    }


@pytest.mark.parametrize(
    ("charmcraft_yaml_template", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - name: test-name
                    channel: test-channel
                actions:
                  pause:
                    description: Pause the database.
                  resume:
                    description: Resume a paused database.
                  {bad_name}:
                    description: Take a snapshot of the database.
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description

                bases:
                  - name: test-name
                    channel: test-channel
                actions:
                  pause:
                    description: Pause the database.
                  resume:
                    description: Resume a paused database.
                  {bad_name}:
                    description: Take a snapshot of the database.
                """
            ),
            None,
        ],
    ],
)
@pytest.mark.parametrize(
    "bad_name",
    [
        "is",
        "-snapshot",
        "111snapshot",
    ],
)
def test_actions_badly_defined_in_charmcraft_yaml(
    tmp_path,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    charmcraft_yaml_template,
    metadata_yaml,
    bad_name,
):
    """test actions badly defined in charmcraft.yaml"""
    prepare_charmcraft_yaml(charmcraft_yaml_template.format(bad_name=bad_name))
    prepare_metadata_yaml(metadata_yaml)

    with pytest.raises(CraftError):
        load(tmp_path)


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - name: test-name
                    channel: test-channel
                actions:
                  pause:
                    description: Pause the database.
                  resume:
                    description: Resume a paused database.
                  snapshot:
                    description: Take a snapshot of the database.
                    params:
                      filename:
                        type: string
                        description: The name of the snapshot file.
                      compression:
                        type: object
                        description: The type of compression to use.
                        properties:
                          kind:
                            type: string
                            enum: [gzip, bzip2, xz]
                          quality:
                            description: Compression quality
                            type: integer
                            minimum: 0
                            maximum: 9
                    required: [filename]
                    additionalProperties: false
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                bases:
                  - name: test-name
                    channel: test-channel
                actions:
                  pause:
                    description: Pause the database.
                  resume:
                    description: Resume a paused database.
                  snapshot:
                    description: Take a snapshot of the database.
                    params:
                      filename:
                        type: string
                        description: The name of the snapshot file.
                      compression:
                        type: object
                        description: The type of compression to use.
                        properties:
                          kind:
                            type: string
                            enum: [gzip, bzip2, xz]
                          quality:
                            description: Compression quality
                            type: integer
                            minimum: 0
                            maximum: 9
                    required: [filename]
                    additionalProperties: false
                """
            ),
            None,
        ],
    ],
)
def test_actions_defined_in_charmcraft_yaml_and_actions_yaml(
    tmp_path,
    create_checker,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    prepare_actions_yaml,
    charmcraft_yaml,
    metadata_yaml,
):
    """actions section cannot be used when actions.yaml file is present."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)
    prepare_actions_yaml(
        dedent(
            """\
            pause:
              description: Pause the database.
            """
        )
    )

    with pytest.raises(CraftError) as cm:
        load(tmp_path)

    assert str(cm.value) == dedent(
        """\
        Bad charmcraft.yaml content:
        - 'actions.yaml' file not allowed when an 'actions' section is defined in 'charmcraft.yaml' in field 'actions'"""
    )


@pytest.mark.parametrize(
    ("charmcraft_yaml", "metadata_yaml"),
    [
        [
            dedent(
                """\
                type: charm
                bases:
                  - name: test-name
                    channel: test-channel
                """
            ),
            dedent(
                """\
                name: test-charm-name-from-metadata-yaml
                summary: test summary
                description: test description
                """
            ),
        ],
        [
            dedent(
                """\
                type: charm
                name: test-charm-name-from-charmcraft-yaml
                summary: test summary
                description: test description
                bases:
                  - name: test-name
                    channel: test-channel
                """
            ),
            None,
        ],
    ],
)
def test_actions_bad_unenforced_defined_in_actions_yaml(
    tmp_path,
    create_checker,
    prepare_charmcraft_yaml,
    prepare_metadata_yaml,
    prepare_actions_yaml,
    charmcraft_yaml,
    metadata_yaml,
):
    """Load a bad actions in actions.yaml. Should not raise an error since check unenforced."""
    prepare_charmcraft_yaml(charmcraft_yaml)
    prepare_metadata_yaml(metadata_yaml)
    prepare_actions_yaml(
        dedent(
            """\
            invalid: 111
            """
        )
    )

    config = load(tmp_path)
    assert config.actions is None



# Copyright 2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Charmcraft-specific code to interface with craft-providers."""

import enum
import os
import pathlib
import sys
from typing import Dict, List, NamedTuple, Optional, Tuple, Union

from craft_cli import CraftError, emit
from craft_providers import Base, Executor, Provider, ProviderError, lxd, multipass
from craft_providers.actions.snap_installer import Snap
from craft_providers.bases import (
    BASE_NAME_TO_BASE_ALIAS,
    get_base_alias,
    get_base_from_alias,
)
from craft_providers.errors import BaseConfigurationError

from charmcraft.bases import check_if_base_matches_host
from charmcraft.env import (
    get_managed_environment_log_path,
    get_managed_environment_snap_channel,
    is_charmcraft_running_from_snap,
    is_charmcraft_running_in_developer_mode,
)
from charmcraft.models.charmcraft import BasesConfiguration
from charmcraft.snap import get_snap_configuration
from charmcraft.utils import confirm_with_user, get_host_architecture


class Plan(NamedTuple):
    """A build plan for a particular base.

    :param bases_config: The BasesConfiguration object, which contains a list of Bases to
      build on and a list of Bases to run on.
    :param build_on: The Base to build on.
    :param bases_index: Index of the BasesConfiguration in bases_config containing the
      Base to build on.
    :param build_on_index: Index of the Base to build on in the BasesConfiguration's build_on list.
    """

    bases_config: BasesConfiguration
    build_on: Base
    bases_index: int
    build_on_index: int


def create_build_plan(
    *,
    bases: Optional[List[BasesConfiguration]],
    bases_indices: Optional[List[int]],
    destructive_mode: bool,
    managed_mode: bool,
    provider: "Provider",
) -> List[Plan]:
    """Determine the build plan based on user inputs and host environment.

    Provide a list of bases that are buildable and scoped according to user
    configuration. Provide all relevant details including the applicable
    bases configuration and the indices of the entries to build for.

    :param bases: List of BaseConfigurations
    :param bases_indices: List of indices of which BaseConfigurations to consider when creating
      the build plan. If None, then all BaseConfigurations are considered
    :param destructive_mode: True is charmcraft is running in destructive mode
    :param managed_mode: True is charmcraft is running in managed mode
    :param provider: Provider object to check for base availability

    :returns: List of Plans
    :raises CraftError: if no bases are provided.
    """
    build_plan: List[Plan] = []

    if not bases:
        raise CraftError("Cannot create build plan because no bases were provided.")

    for bases_index, bases_config in enumerate(bases):
        if bases_indices and bases_index not in bases_indices:
            emit.debug(f"Skipping 'bases[{bases_index:d}]' due to --base-index usage.")
            continue

        for build_on_index, build_on in enumerate(bases_config.build_on):
            if managed_mode or destructive_mode:
                matches, reason = check_if_base_matches_host(build_on)
            else:
                matches, reason = is_base_available(build_on)

            if matches:
                emit.debug(
                    f"Building for 'bases[{bases_index:d}]' "
                    f"as host matches 'build-on[{build_on_index:d}]'.",
                )
                build_plan.append(Plan(bases_config, build_on, bases_index, build_on_index))
                break
            else:
                emit.progress(
                    f"Skipping 'bases[{bases_index:d}].build-on[{build_on_index:d}]': "
                    f"{reason}.",
                )
        else:
            emit.progress(
                "No suitable 'build-on' environment found "
                f"in 'bases[{bases_index:d}]' configuration.",
                permanent=True,
            )

    return build_plan


def get_command_environment(base: Base) -> Dict[str, str]:
    """Construct the required environment."""
    env = base.default_command_environment()
    env["CHARMCRAFT_MANAGED_MODE"] = "1"

    # Pass-through host environment that target may need.
    for env_key in ["http_proxy", "https_proxy", "no_proxy"]:
        if env_key in os.environ:
            env[env_key] = os.environ[env_key]

    return env


def get_instance_name(
    *,
    bases_index: int,
    build_on_index: int,
    project_name: str,
    project_path: pathlib.Path,
    target_arch: str,
) -> str:
    """Formulate the name for an instance using each of the given parameters.

    Incorporate each of the parameters into the name to come up with a
    predictable naming schema that avoids name collisions across multiple,
    potentially complex, projects.

    :param bases_index: Index of `bases:` entry.
    :param build_on_index: Index of `build-on` within bases entry.
    :param project_name: Name of charm project.
    :param project_path: Directory of charm project.
    :param target_arch: Targeted architecture, used in the name to prevent
        collisions should future work enable multiple architectures on the same
        platform.
    """
    return "-".join(
        [
            "charmcraft",
            project_name,
            str(project_path.stat().st_ino),
            str(bases_index),
            str(build_on_index),
            target_arch,
        ]
    )


def get_base_configuration(
    *,
    alias: enum.Enum,
    instance_name: str,
    shared_cache_path: Optional[pathlib.Path] = None,
) -> Base:
    """Create a Base configuration."""
    # injecting a snap on a non-linux system is not supported, so default to
    # install charmcraft from the store's stable channel
    snap_channel = get_managed_environment_snap_channel()
    if snap_channel is None and sys.platform != "linux":
        snap_channel = "stable"

    base = get_base_from_alias(alias)
    charmcraft_snap = Snap(name="charmcraft", channel=snap_channel, classic=True)
    environment = get_command_environment(base)
    return base(
        alias=alias,
        environment=environment,
        hostname=instance_name,
        snaps=[charmcraft_snap],
        compatibility_tag=f"charmcraft-{base.compatibility_tag}.0",
        cache_path=shared_cache_path,
    )


def capture_logs_from_instance(instance: Executor) -> None:
    """Retrieve logs from instance.

    :param instance: Instance to retrieve logs from.
    """
    source_log_path = get_managed_environment_log_path()
    with instance.temporarily_pull_file(source=source_log_path, missing_ok=True) as local_log_path:
        if local_log_path:
            emit.debug("Logs captured from managed instance:")
            with open(local_log_path, encoding="utf8") as fh:
                for line in fh:
                    emit.debug(f":: {line.rstrip()}")
        else:
            emit.debug("No logs found in instance.")
            return


def ensure_provider_is_available(provider: "Provider") -> None:
    """Ensure provider is installed, running, and properly configured.

    If the provider is not installed, the user is prompted to install it.

    :param instance: the provider to ensure is available
    :raises ProviderError: if provider is not available, or if the user
    chooses not to install the provider.
    """
    # TODO: add provider.name and provider.install_recommendations to craft-providers
    confirm_msg = (
        "Provider is required but not installed. Do you wish to "
        "install provider and configure it with the defaults?"
    )
    if not provider.is_provider_installed() and not confirm_with_user(confirm_msg, default=False):
        raise ProviderError("Provider is required, but not installed.")
    provider.ensure_provider_is_available()


def is_base_available(base: Base) -> Tuple[bool, Union[str, None]]:
    """Check if provider can provide an environment matching given base.

    :param base: Base to check.

    :returns: Tuple of bool indicating whether it is a match, with optional
            reason if not a match.
    """
    arch = get_host_architecture()
    if arch not in base.architectures:
        return (
            False,
            f"host architecture {arch!r} not in base architectures {base.architectures!r}",
        )

    if base.name not in ("ubuntu", "centos", "almalinux"):
        return (
            False,
            f"name {base.name!r} is not yet supported (must be 'ubuntu', 'almalinux', "
            "or 'centos')",
        )

    try:
        get_base_alias((base.name, base.channel))
    except BaseConfigurationError:
        *firsts, last = sorted(" ".join(s) for s in BASE_NAME_TO_BASE_ALIAS)
        allowed = f"{', '.join(map(repr, firsts))} or {last!r}"
        return (
            False,
            f"base {base.name!r} channel {base.channel!r} is not yet supported "
            f"(must be {allowed})",
        )

    return True, None


def _get_platform_default_provider() -> str:
    if sys.platform == "linux":
        return "lxd"

    return "multipass"


def get_provider():
    """Get the configured or appropriate provider for the host OS.

    If platform is not Linux, use Multipass.

    If platform is Linux:
    (1) use provider specified with CHARMCRAFT_PROVIDER if running
        in developer mode,
    (2) use provider specified with snap configuration if running
        as snap,
    (3) default to platform default (LXD on Linux).

    :return: Provider instance.
    """
    provider = None

    if is_charmcraft_running_in_developer_mode():
        provider = os.getenv("CHARMCRAFT_PROVIDER")

    if provider is None and is_charmcraft_running_from_snap():
        snap_config = get_snap_configuration()
        provider = snap_config.provider if snap_config else None

    if provider is None:
        provider = _get_platform_default_provider()

    if provider == "lxd":
        return lxd.LXDProvider(lxd_project="charmcraft")
    elif provider == "multipass":
        return multipass.MultipassProvider()

    raise CraftError(f"Unsupported provider specified {provider!r}.")



# Copyright 2021-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Analyze and lint charm structures and files."""
import abc
import ast
import os
import pathlib
import shlex
from typing import Generator, List, Type, Union

import yaml

from charmcraft import config, utils
from charmcraft.metafiles.metadata import parse_charm_metadata_yaml, read_metadata_yaml
from charmcraft.models.lint import CheckResult, CheckType, LintResult

# the documentation page for "Analyzers and linters"
BASE_DOCS_URL = "https://juju.is/docs/sdk/charmcraft-analyzers-and-linters"


def get_entrypoint_from_dispatch(basedir: pathlib.Path) -> Union[pathlib.Path, None]:
    """Verify if the charm has a dispatch file pointing to a Python entrypoint.

    :returns: the entrypoint path if all succeeds, None otherwise.
    """
    # get the entrypoint from the last useful dispatch line
    dispatch = basedir / "dispatch"
    entrypoint_str = ""
    try:
        with dispatch.open("rt", encoding="utf8") as fh:
            last_line = None
            for line in fh:
                if line.strip():
                    last_line = line
            if last_line:
                entrypoint_str = shlex.split(last_line)[-1]
    except (OSError, UnicodeDecodeError):
        return None
    if not entrypoint_str:
        return None
    return basedir / entrypoint_str


def check_dispatch_with_python_entrypoint(basedir: pathlib.Path) -> Union[pathlib.Path, None]:
    """Verify if the charm has a dispatch file pointing to a Python entrypoint.

    :returns: the entrypoint path if all succeeds, None otherwise.
    """
    entrypoint = get_entrypoint_from_dispatch(basedir)
    if entrypoint and entrypoint.suffix == ".py" and os.access(entrypoint, os.X_OK):
        return entrypoint
    return None


class BaseChecker(metaclass=abc.ABCMeta):
    """Base class for checker classes."""

    check_type: CheckType
    name: str
    url: str
    text: str

    @abc.abstractmethod
    def run(self, basedir: pathlib.Path) -> str:
        """Run this checker."""
        ...


class AttributeChecker(BaseChecker, metaclass=abc.ABCMeta):
    """Base attribute checker."""

    check_type = CheckType.ATTRIBUTE


class Linter(BaseChecker, metaclass=abc.ABCMeta):
    """Base linter class."""

    check_type = CheckType.LINT
    Result = LintResult


class Language(AttributeChecker):
    """Check the language used to write the charm.

    Currently only Python is detected, if the following checks are true:

    - the charm has a text dispatch with a python call
    - the charm has a `.py` entry point
    - the entry point file is executable
    """

    name = "language"
    url = BASE_DOCS_URL + "#heading--language"
    text = "The charm is written with Python."

    class Result:
        """Possible results for this attribute checker."""

        UNKNOWN = LintResult.UNKNOWN
        PYTHON = "python"

    def run(self, basedir: pathlib.Path) -> str:
        """Run the proper verifications."""
        python_entrypoint = check_dispatch_with_python_entrypoint(basedir)
        return self.Result.UNKNOWN if python_entrypoint is None else self.Result.PYTHON


class Framework(AttributeChecker):
    """Check the framework the charm is based on.

    Currently it detects if the Operator Framework is used, if...

    - the language attribute is set to python
    - the charm contains venv/ops
    - the charm imports ops in the entry point.

    ...or the Reactive Framework is used, if the charm...

    - has a metadata.yaml with "name" in it
    - has a reactive/<name>.py file that imports "charms.reactive"
    - has a file name that starts with "charms.reactive-" inside the "wheelhouse" directory
    """

    name = "framework"
    url = BASE_DOCS_URL + "#heading--framework"

    class Result:
        """Possible results for this attribute checker."""

        OPERATOR = "operator"
        REACTIVE = "reactive"
        UNKNOWN = LintResult.UNKNOWN

    # different texts to be exposed as `text` (see the property below)
    result_texts = {
        Result.OPERATOR: "The charm is based on the Operator Framework.",
        Result.REACTIVE: "The charm is based on the Reactive Framework.",
        Result.UNKNOWN: "The charm is not based on any known Framework.",
    }

    def __init__(self):
        self.result = None

    @property
    def text(self):
        """Return a text in function of the result state."""
        if self.result is None:
            return None
        return self.result_texts[self.result]

    def _get_imports(self, filepath: pathlib.Path) -> Generator[List[str], None, None]:
        """Parse a Python filepath and yield its imports.

        If the file does not exist or cannot be parsed, return empty. Otherwise
        return the name for each imported module, split by possible dots.
        """
        if not os.access(filepath, os.R_OK):
            return
        try:
            parsed = ast.parse(filepath.read_bytes())
        except SyntaxError:
            return

        for node in ast.walk(parsed):
            if isinstance(node, ast.Import):
                for name in node.names:
                    yield name.name.split(".")
            elif isinstance(node, ast.ImportFrom):
                yield node.module.split(".")

    def _check_operator(self, basedir: pathlib.Path) -> bool:
        """Detect if the Operator Framework is used."""
        python_entrypoint = check_dispatch_with_python_entrypoint(basedir)
        if python_entrypoint is None:
            return False

        opsdir = basedir / "venv" / "ops"
        if not opsdir.exists() or not opsdir.is_dir():
            return False

        for import_parts in self._get_imports(python_entrypoint):
            if import_parts[0] == "ops":
                return True
        return False

    def _check_reactive(self, basedir: pathlib.Path) -> bool:
        """Detect if the Reactive Framework is used."""
        try:
            metadata = parse_charm_metadata_yaml(basedir)
        except Exception:
            # file not found, corrupted, or mandatory "name" not present
            return False

        wheelhouse_dir = basedir / "wheelhouse"
        if not wheelhouse_dir.exists():
            return False
        if not any(f.name.startswith("charms.reactive-") for f in wheelhouse_dir.iterdir()):
            return False

        module_basename = metadata.name.replace("-", "_")
        entrypoint = basedir / "reactive" / f"{module_basename}.py"
        for import_parts in self._get_imports(entrypoint):
            if import_parts[0] == "charms" and import_parts[1] == "reactive":
                return True
        return False

    def run(self, basedir: pathlib.Path) -> str:
        """Run the proper verifications."""
        if self._check_operator(basedir):
            result = self.Result.OPERATOR
        elif self._check_reactive(basedir):
            result = self.Result.REACTIVE
        else:
            result = self.Result.UNKNOWN
        self.result = result
        return result


class JujuMetadata(Linter):
    """Check that the metadata.yaml file exists and is valid.

    The charm is considered to have a valid metadata if the following checks are true:

    - the metadata.yaml is present
    - it is a valid YAML file
    - it has at least the following fields: name, summary, and description
    """

    name = "metadata"
    url = BASE_DOCS_URL + "#heading--metadata"

    def __init__(self):
        self.text = ""

    def run(self, basedir: pathlib.Path) -> str:
        """Run the proper verifications."""
        try:
            metadata = read_metadata_yaml(basedir)
        except yaml.YAMLError:
            self.text = "The metadata.yaml file is not a valid YAML file."
            return self.Result.ERRORS
        except Exception:
            self.text = "Cannot read the metadata.yaml file."
            return self.Result.ERRORS

        # check required attributes
        missing_fields = {"name", "summary", "description"} - set(metadata)
        if missing_fields:
            missing = utils.humanize_list(missing_fields, "and")
            self.text = f"The metadata.yaml file is missing the following attribute(s): {missing}."
            return self.Result.ERRORS

        return self.Result.OK


class JujuActions(Linter):
    """Check that the actions.yaml file is valid YAML if it exists."""

    name = "juju-actions"
    url = BASE_DOCS_URL + "#heading--juju-actions"
    text = "The actions.yaml file is not a valid YAML file."

    def run(self, basedir: pathlib.Path) -> str:
        """Run the proper verifications."""
        filepath = basedir / "actions.yaml"
        if not filepath.exists():
            # it's optional
            return self.Result.OK

        try:
            with filepath.open("rt", encoding="utf8") as fh:
                yaml.safe_load(fh)
        except Exception:
            return self.Result.ERRORS

        return self.Result.OK


class JujuConfig(Linter):
    """Check that the config.yaml file (if it exists) is valid.

    The file is considered valid if the following checks are true:

    - has an 'options' key
    - it is a dictionary
    - each item inside has the mandatory 'type' key
    """

    name = "juju-config"
    url = BASE_DOCS_URL + "#heading--juju-config"

    def __init__(self):
        self.text = ""

    def run(self, basedir: pathlib.Path) -> str:
        """Run the proper verifications."""
        filepath = basedir / "config.yaml"
        if not filepath.exists():
            # it's optional
            return self.Result.OK

        try:
            with filepath.open("rt", encoding="utf8") as fh:
                content = yaml.safe_load(fh)
        except Exception:
            self.text = "The config.yaml file is not a valid YAML file."
            return self.Result.ERRORS

        options = content.get("options")
        if not isinstance(options, dict):
            self.text = "Error in config.yaml: must have an 'options' dictionary."
            return self.Result.ERRORS

        for value in options.values():
            if "type" not in value:
                self.text = "Error in config.yaml: items under 'options' must have a 'type' key."
                return self.Result.ERRORS

        return self.Result.OK


class Entrypoint(Linter):
    """Check the entrypoint is correct.

    It validates that the entrypoint, if used by 'dispatch', ...

    - exists
    - is a file
    - is executable
    """

    name = "entrypoint"
    url = BASE_DOCS_URL + "#heading--entrypoint"

    def __init__(self):
        self.text = ""

    def run(self, basedir: pathlib.Path) -> str:
        """Run the proper verifications."""
        entrypoint = get_entrypoint_from_dispatch(basedir)
        if entrypoint is None:
            self.text = "Cannot find a proper 'dispatch' script pointing to an entrypoint."
            return self.Result.NONAPPLICABLE

        if not entrypoint.exists():
            self.text = f"Cannot find the entrypoint file: {str(entrypoint)!r}"
            return self.Result.ERRORS

        if not entrypoint.is_file():
            self.text = f"The entrypoint is not a file: {str(entrypoint)!r}"
            return self.Result.ERRORS

        if not os.access(entrypoint, os.X_OK):
            self.text = f"The entrypoint file is not executable: {str(entrypoint)!r}"
            return self.Result.ERRORS

        return self.Result.OK


# all checkers to run; the order here is important, as some checkers depend on the
# results from others
CHECKERS: List[Type[BaseChecker]] = [
    Language,
    JujuActions,
    JujuConfig,
    JujuMetadata,
    Framework,
    Entrypoint,
]


def analyze(
    config: config.CharmcraftConfig,
    basedir: pathlib.Path,
    *,
    override_ignore_config: bool = False,
) -> List[CheckResult]:
    """Run all checkers and linters."""
    all_results = []
    for cls in CHECKERS:
        # do not run the ignored ones
        if cls.check_type == CheckType.ATTRIBUTE:
            ignore_list = config.analysis.ignore.attributes
        else:
            ignore_list = config.analysis.ignore.linters
        if cls.name in ignore_list and not override_ignore_config:
            all_results.append(
                CheckResult(
                    check_type=cls.check_type,
                    name=cls.name,
                    result=LintResult.IGNORED,
                    url=cls.url,
                    text="",
                )
            )
            continue

        checker = cls()
        try:
            result = checker.run(basedir)
        except Exception:
            result = (
                LintResult.UNKNOWN
                if checker.check_type == CheckType.ATTRIBUTE
                else LintResult.FATAL
            )
        all_results.append(
            CheckResult(
                check_type=checker.check_type,
                name=checker.name,
                url=checker.url,
                text=checker.text or "n/a",
                result=result,
            )
        )
    return all_results
